/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/*
 * FML interpreter. Europagate, 1995
 *
 * $Log: fmltoken.c,v $
 * Revision 1.7  1995/05/16 09:39:35  adam
 * LICENSE.
 *
 * Revision 1.6  1995/02/23  08:32:07  adam
 * Changed header.
 *
 * Revision 1.4  1995/02/10  18:15:53  adam
 * FML function 'strcmp' implemented. This function can be used to
 * test for existence of MARC fields.
 *
 * Revision 1.3  1995/02/10  15:50:57  adam
 * MARC interface implemented. Minor bugs fixed. fmltest can
 * be used to format single MARC records. New function '\list'
 * implemented.
 *
 * Revision 1.2  1995/02/07  16:09:24  adam
 * The \ character is no longer INCLUDED when terminating a token.
 * Major changes in tokenization routines. Bug fixes in expressions
 * with lists (fml_sub0).
 *
 * Revision 1.1.1.1  1995/02/06  13:48:10  adam
 * First version of the FML interpreter. It's slow and memory isn't
 * freed properly. In particular, the FML nodes aren't released yet.
 *
 */
#include <string.h>
#include <assert.h>
#include <stdio.h>

#include "fmlp.h"

static int look_char;
static int look_type;
static char lex_buf[FML_MAX_TOKEN];

static void lexer (Fml fml);

struct fml_node *fml_group (Fml fml);

struct fml_node *fml_tokenize (Fml fml)
{
    struct fml_node *p;

    look_char = (*fml->read_func)();
    p = fml_group (fml);
    if (fml->debug)
    {
        fml_pr_list (p);
        printf ("\n");
    }
    if (look_char != fml->eof_mark)
    {
        fprintf (stderr, "Ill formed parantheses");
        exit (1);
    }
    return p;
}

void fml_pr_list (struct fml_node *p)
{
    printf ("{");

    while (p)
    {
        if (p->is_atom)
        {
            char buf[100];
            fml_atom_str (p->p[0], buf);
            printf (" %s", buf);
        }
        else
        {
            printf (" ");
            fml_pr_list (p->p[0]);
        }
        p = p->p[1];
    }
    printf (" }");         
}

struct fml_node *fml_group (Fml fml)
{
    struct fml_node *ptr0 = NULL, *ptr1, *ptr2;

    lexer (fml);
    if (look_type == 0)
        return NULL;
    while (1)
    {
        if (look_type == 'a')
        {
            ptr2 = fml_node_alloc (fml);
            if (!ptr0)
                ptr0 = ptr2;
            else
                ptr1->p[1] = ptr2;
            ptr2->p[0] = fml_atom_alloc (fml, lex_buf);
            ptr2->is_atom = 1;
        }
        else if (look_type == '{')
        {
            struct fml_node *sptr = fml_group (fml);
            if (sptr)
            {
                ptr2 = fml_node_alloc (fml);
                if (!ptr0)
                    ptr0 = ptr2;
                else
                        ptr1->p[1] = ptr2;
                ptr2->p[0] = sptr;
                ptr2->is_atom = 0;
            }
            else
            {
                ptr2 = fml_node_alloc (fml);
                if (!ptr0)
                    ptr0 = ptr2;
                else
                    ptr1->p[1] = ptr2;
                ptr2->is_atom = 0;
            }
        }
        else
            break;
        lexer (fml);
        ptr1 = ptr2;
    }
    return ptr0;
}

static void lexer (Fml fml)
{
    int off;
    while (1) 
    {
        if (look_char == fml->eof_mark)
        {
            look_type = 0;
            return;
        }
        else if (look_char == fml->comment_char)
        {
            do
                look_char = (*fml->read_func)();
            while (look_char != '\n' && look_char != fml->eof_mark);
        }
        else
        {
            if (!strchr (fml->white_chars, look_char))
                break;
            look_char = (*fml->read_func)();
        }
    }
    if (look_char == '{')
    {
        look_type = '{';
        look_char = (*fml->read_func)();
    }
    else if (look_char == '}')
    {
        look_type = '}';
        look_char = (*fml->read_func)();
    }        
    else if (look_char == '\'')
    {
        off = 0;
        look_char = (*fml->read_func)();
        while (look_char != fml->eof_mark && look_char != '\'')
        {
            lex_buf[off++] = look_char;
            if (look_char == '\\')
            {
                look_char = (*fml->read_func)();
                switch (look_char)
                {
                case 'n':
                    lex_buf[off-1] = '\n';
                    break;
                case 't':
                    lex_buf[off-1] = '\n';
                    break;
                case '\'':
                    lex_buf[off-1] = '\''; 
                    break;
                default:
                    lex_buf[off-1] = look_char;
                }
            }
            look_char = (*fml->read_func)();
        } 
        lex_buf[off] = '\0';
        look_type = 'a';
        if (look_char == '\'')
            look_char = (*fml->read_func)();
    }
    else
    {
        off = 0;
        do
        {
            lex_buf[off++] = look_char;
            look_char = (*fml->read_func)();
        } while (look_char != fml->eof_mark
                 && !strchr (fml->white_chars, look_char)
                 && look_char != '{' && look_char != '}');
        lex_buf[off] = '\0';
        look_type = 'a';
    }
#if 0
    if (fml->debug)
    {
        if (look_type == 'a')
            printf ("[%s]", lex_buf);
        else
            printf ("[%c]", look_type);
    }
#endif
}
