/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/*
 * Record management - public header file
 *
 * $Log: iso2709.h,v $
 * Revision 1.11  1995/05/16 09:39:39  adam
 * LICENSE.
 *
 * Revision 1.10  1995/03/30  07:32:46  adam
 * New log-level: GW_LOG_NOTICE. New 2709 function: iso2709_mk.
 *
 * Revision 1.9  1995/03/29  11:44:25  adam
 * New functions: iso2709_a_.. for record manipulation.
 *
 * Revision 1.8  1995/03/28  16:06:42  adam
 * New function: iso2709_out.
 *
 * Revision 1.7  1995/03/27  12:50:40  adam
 * New level: GW_LOG_ERRNO.
 *
 * Revision 1.6  1995/03/08  12:36:35  adam
 * New function: dbc2709_cvt.
 *
 * Revision 1.5  1995/02/23  08:32:12  adam
 * Changed header.
 *
 * Revision 1.3  1995/02/10  17:05:14  adam
 * New function iso2709_display to display MARC records in a
 * line-by-line format. The iso2709_cvt function no longer
 * prints the record to stderr.
 *
 * Revision 1.2  1995/02/10  16:50:25  adam
 * Indicator field moved to 'struct iso2709_dir' from 'struct
 * iso2709_field'.
 * Function iso2709_rm implemented - to delete a MARC record.
 *
 * Revision 1.1.1.1  1995/02/09  17:27:12  adam
 * Initial version of email gateway under CVS control.
 *
 */

#ifndef ISO2709_H
#define ISO2709_H

typedef struct iso2709_rec *Iso2709Rec;
typedef struct iso2709_anchor *Iso2709Anchor;

char *iso2709_read (FILE *inf);
Iso2709Rec iso2709_cvt (const char *buf);
Iso2709Rec iso2709_mk (void);
void iso2709_rm (Iso2709Rec rec);
void iso2709_display (Iso2709Rec rec, FILE *out);
int iso2709_out (Iso2709Rec p, char **buf, int size);

Iso2709Anchor iso2709_a_mk (Iso2709Rec rec);
void iso2709_a_rm (Iso2709Anchor anchor);
int iso2709_a_first (Iso2709Anchor anchor);
int iso2709_a_next_line (Iso2709Anchor anchor);
int iso2709_a_next_field (Iso2709Anchor anchor);
int iso2709_a_next (Iso2709Anchor anchor);
int iso2709_a_info_field (Iso2709Anchor anchor,
                    char **tag, char **indicator,
                    char **identifier, char **data);
int iso2709_a_info_line (Iso2709Anchor anchor,
                    char **tag, char **indicator);
int iso2709_a_delete_field (Iso2709Anchor anchor);
int iso2709_a_delete_line (Iso2709Anchor anchor);
int iso2709_a_insert (Iso2709Anchor anchor, 
                      const char *tag, const char *indicator,
                      const char *identifier, const char *data);
int iso2709_a_search (Iso2709Anchor anchor,
                      const char *tag_p, const char *indicator_p,
                      const char *identifier_p);
#endif

