/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/* Gateway tty print utility
 * Europagate, 1995
 *
 * $Log: ttyemit.c,v $
 * Revision 1.2  1995/05/16 09:40:55  adam
 * LICENSE.
 *
 * Revision 1.1  1995/04/17  09:36:34  adam
 * ttyemit moved from kernel directory.
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <ttyemit.h>

static char line_buf[256];
static int  line_col = 0;
static int  esc_flag = 0;

static int  line_min = 30;
static int  line_max = 76;
static FILE *out_f = stdout;

void tty_init (FILE *out, int min, int max)
{
    out_f = out;
    if (min > 0)
        line_min = min;
    if (max > 2)
        line_max = max;
    line_col = 0;
    esc_flag = 0;
}

static void flush (void)
{
    int j;

    for (j = 0; j<line_col; j++)
        putc (line_buf[j], out_f);
    putc ('\n', out_f);
    line_col = 0;
}

static void split (int ch)
{
    int i = line_col, j;

    while (1)
        if (line_buf[--i] == ' ')
        {
            int extra = 0;

            for (j = 0; j<i; j++)
            {
#if 1
                if (j>1 && line_buf[j] >= 'A'&& line_buf[j] <= 'Z'
                     && line_buf[j-1] == ' '
                     && (line_buf[j-2] == '.' || line_buf[j-2] == ';')
                     && extra < line_max-i )
                {
                    extra++;
                    putc (' ', out_f);
                }
#endif
                putc (line_buf[j], out_f);
            }
            putc ('\n', out_f);
            j = 0;
            ++i;
            while (i < line_col)
                line_buf[j++] = line_buf[i++];
            line_col = j;
            break;
        }
        else if (i < line_min)
            flush ();
}

static void escape (int ch)
{
    switch (ch)
    {
    case 'n':
	if (line_col >= line_max)
	    split (' ');
	else
            flush ();
        break;
    case 't':
        do
            line_buf[line_col++] = ' ';
        while (line_col & 7);
        break;
    case 's':
        line_buf[line_col++] = ' ';
        break;
    default:
        line_buf[line_col++] = ch;
    }
}

void tty_emit (int ch)
{
    if (esc_flag)
    {
        escape (ch);
        esc_flag = 0;
    }
    else if (ch == '\\')
        esc_flag = 1;
    else if (ch == '\n')
        flush ();
    else if (line_col || ch != ' ')
    {
        line_buf[line_col++] = ch;
        if (line_col >= line_max)
            split (ch);
    }
}
