/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $Log: wsh.c,v $
 * Revision 1.13  1996/02/29 15:36:34  adam
 * Main shell loop implemented as a separate function called wshmain.
 *
 * Revision 1.12  1996/02/19  15:45:00  adam
 * Sort of targets in multiple-targets selection.
 *
 * Revision 1.11  1996/02/12  10:10:32  adam
 * Resource/config system used by the gateway.
 *
 * Revision 1.10  1996/01/12  10:05:22  adam
 * If script name ends with ';' HTTP/GET/Expires will be defined.
 * The cgi interface only reads final handshake if response from
 * server (shell) was zero-terminated [If it isn't it probably died].
 *
 * Revision 1.9  1996/01/09  10:46:51  adam
 * New defines: LOGDIR/EGWDIR/CGIDIR set in Makefile.
 *
 * Revision 1.8  1995/12/20  16:31:33  adam
 * Bug fix: shell might terminate even though new request was initiated
 * by the cgi interface program.
 * Work on more simple user interface and Europagate buttons.
 *
 * Revision 1.7  1995/11/06  17:44:23  adam
 * State reestablised when shell restarts. History of previous
 * result sets.
 *
 * Revision 1.6  1995/11/06  10:51:19  adam
 * End of response marker in response from wsh/wproto to wcgi.
 * Shells are respawned when necessary.
 *
 * Revision 1.5  1995/11/01  16:15:48  adam
 * Better presentation of records. Query/set number persistent.
 *
 * Revision 1.4  1995/10/27  15:12:12  adam
 * IrTcl incorporated in the gateway.
 * Better separation of script types.
 * Z39.50 gateway scripts entered.
 *
 * Revision 1.3  1995/10/23  16:55:41  adam
 * A lot of changes - really.
 *
 * Revision 1.2  1995/10/20  14:02:42  adam
 * First version of WWW gateway with embedded Tcl.
 *
 * Revision 1.1  1995/10/20  11:49:28  adam
 * First version of www gateway.
 *
 */

#include <stdio.h>
#include "whtml.h"
#include "wtcl.h"
#include "wirtcl.h"

int main (int argc, char **argv)
{
    W_Interp_Type w_interp_type;

    if (!strcmp (*argv, "egwtcl"))
        w_interp_type = w_interp_tcl;
    else if (!strcmp (*argv, "egwirtcl"))
        w_interp_type = w_interp_irtcl;
    else if (!strcmp (*argv, "egwhtml"))
        w_interp_type = w_interp_html;
    else
    {
        gw_log (GW_LOG_FATAL, "wsh", "Cannot determine shell type. prog=%s",
	        *argv);
        exit (1);
    }
    wshmain (argc, argv, w_interp_type);
    exit (0);
}


